from typing_extensions import Annotated, TypedDict

from .generate.type_wrapper import Field, add_extra_schema


class StreamRange(TypedDict):
    """The parameters required to describe a sequence of incrementing integers"""

    start: Annotated[
        int,
        Field(description="First number in the range"),
    ]
    stop: Annotated[
        int,
        Field(description="Last number in the range is less than this number"),
    ]


STREAM_DATUM_EXTRA_SCHEMA = {"additionalProperties": False}


@add_extra_schema(STREAM_DATUM_EXTRA_SCHEMA)
class StreamDatum(TypedDict):
    """Document to reference a quanta of an externally-stored stream of data."""

    descriptor: Annotated[
        str,
        Field(description="UID of the EventDescriptor to " "which this Datum belongs"),
    ]
    stream_resource: Annotated[
        str,
        Field(
            description="The UID of the Stream Resource to which this Datum belongs."
        ),
    ]
    uid: Annotated[
        str,
        Field(
            description="Globally unique identifier for this Datum. A suggested "
            "formatting being '<stream_resource>/<stream_name>/<block_id>"
        ),
    ]
    seq_nums: Annotated[
        StreamRange,
        Field(
            description="A slice object showing the Event numbers the "
            "resource corresponds to"
        ),
    ]
    indices: Annotated[
        StreamRange,
        Field(
            description="A slice object passed to the StreamResource "
            "handler so it can hand back data and timestamps"
        ),
    ]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from pathlib import Path
from subprocess import check_output

import requests

import event_model

# -- General configuration ------------------------------------------------

# General information about the project.
project = "event-model"
copyright = "2019, Brookhaven National Lab"
author = "Brookhaven National Lab"


# The full version, including alpha/beta/rc tags.
release = event_model.__version__

# The short X.Y version.
if "+" in release:
    # Not on a tag, use branch name
    root = Path(__file__).absolute().parent.parent
    git_branch = check_output("git branch --show-current".split(), cwd=root)
    version = git_branch.decode().strip()
else:
    version = release

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

extensions = [
    # Use this for generating API docs
    "sphinx.ext.autodoc",
    # This can parse google style docstrings
    "sphinx.ext.napoleon",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
    # Adds the inheritance-diagram generation directive
    "sphinx.ext.inheritance_diagram",
    # Add a copy button to each code block
    "sphinx_copybutton",
    # For the card element
    "sphinx_design",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "matplotlib.sphinxext.plot_directive",
    "sphinx_copybutton",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

napoleon_google_docstring = False
napoleon_numpy_docstring = True

# If true, Sphinx will warn about all references where the target cannot
# be found.
nitpicky = False

# A list of (type, target) tuples (by default empty) that should be ignored when
# generating warnings in "nitpicky mode". Note that type should include the
# domain name if present. Example entries would be ('py:func', 'int') or
# ('envvar', 'LD_LIBRARY_PATH').
nitpick_ignore = [
    ("py:class", "NoneType"),
    ("py:class", "'str'"),
    ("py:class", "'float'"),
    ("py:class", "'int'"),
    ("py:class", "'bool'"),
    ("py:class", "'object'"),
    ("py:class", "'id'"),
    ("py:class", "typing_extensions.Literal"),
]

# Both the class’ and the __init__ method’s docstring are concatenated and
# inserted into the main body of the autoclass directive
autoclass_content = "both"

# Order the members by the order they appear in the source code
autodoc_member_order = "bysource"

# Don't inherit docstrings from baseclasses
autodoc_inherit_docstrings = False

# Output graphviz directive produced images in a scalable format
graphviz_output_format = "svg"

# The name of a reST role (builtin or Sphinx extension) to use as the default
# role, that is, for text marked up `like this`
default_role = "any"

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# These patterns also affect html_static_path and html_extra_path
exclude_patterns = ["_build"]

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# Example configuration for intersphinx: refer to the Python standard library.
# This means you can link things like `str` and `asyncio` to the relevant
# docs in the python documentation.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "cachetools": ("https://cachetools.readthedocs.io/en/stable/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "matplotlib": ("https://matplotlib.org", None),
    "jsonschema": ("https://python-jsonschema.readthedocs.io/en/stable/", None),
}

# A dictionary of graphviz graph attributes for inheritance diagrams.
inheritance_graph_attrs = dict(rankdir="TB")

# Common links that should be available on every page
rst_epilog = """
.. _NSLS: https://www.bnl.gov/nsls2
.. _black: https://github.com/psf/black
.. _flake8: https://flake8.pycqa.org/en/latest/
.. _isort: https://github.com/PyCQA/isort
.. _mypy: http://mypy-lang.org/
.. _pre-commit: https://pre-commit.com/
"""

# Ignore localhost links for periodic check that links in docs are valid
linkcheck_ignore = [r"http://localhost:\d+/"]

# Set copy-button to ignore python and bash prompts
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for manual page output ---------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "event-model.tex",
        "Bluesky Event Model Documentation",
        "Contributors",
        "manual",
    ),
]


# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "event-model", "Bluesky Event Model Documentation", [author], 1)
]

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "event-model",
        "Bluesky Event Model Documentation",
        author,
        "event-model",
        "Data model used by the bluesky ecosystem",
        "Miscellaneous",
    ),
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
github_repo = project
github_user = "bluesky"
switcher_json = f"https://{github_user}.github.io/{github_repo}/switcher.json"
switcher_exists = requests.get(switcher_json).ok
if not switcher_exists:
    print(
        "*** Can't read version switcher, is GitHub pages enabled? \n"
        "    Once Docs CI job has successfully run once, set the "
        "Github pages source branch to be 'gh-pages' at:\n"
        f"    https://github.com/{github_user}/{github_repo}/settings/pages",
        file=sys.stderr,
    )

# Theme options for pydata_sphinx_theme
# We don't check switcher because there are 3 possible states for a repo:
# 1. New project, docs are not published so there is no switcher
# 2. Existing project with latest skeleton, switcher exists and works
# 3. Existing project with old skeleton that makes broken switcher,
#    switcher exists but is broken
# Point 3 makes checking switcher difficult, because the updated skeleton
# will fix the switcher at the end of the docs workflow, but never gets a chance
# to complete as the docs build warns and fails.
html_theme_options = dict(
    logo=dict(
        text=project,
    ),
    use_edit_page_button=True,
    github_url=f"https://github.com/{github_user}/{github_repo}",
    icon_links=[
        dict(
            name="PyPI",
            url=f"https://pypi.org/project/{project}",
            icon="fas fa-cube",
        )
    ],
    switcher=dict(
        json_url=switcher_json,
        version_match=version,
    ),
    check_switcher=False,
    navbar_end=["theme-switcher", "icon-links", "version-switcher"],
)

# A dictionary of values to pass into the template engine’s context for all pages
html_context = dict(
    github_user=github_user,
    github_repo=project,
    github_version=version,
    doc_path="docs",
)

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Logo
html_logo = "images/bluesky-logo-dark.svg"
# html_favicon = "images/dls-favicon.ico"



from uuid import uuid4

import jsonschema
import pytest
from jsonschema.exceptions import ValidationError

import event_model
from event_model import DocumentNames, schema_validators


def new_uid():
    return str(uuid4())


def test_data_session():
    run_bundle = event_model.compose_run(
        uid="42",
        metadata={"data_groups": ["a", "b"], "data_session": "magrathia_visit_42"},
    )
    start_doc = run_bundle.start_doc
    assert start_doc["data_groups"] == ["a", "b"]

    with pytest.raises(ValidationError):
        event_model.compose_run(
            uid="42", metadata={"data_session": ["shoud", "not", "be", "a", "list"]}
        )

    with pytest.raises(ValidationError):
        event_model.compose_run(
            uid="42", metadata={"data_groups": "this should be an array"}
        )


def test_dots_not_allowed_in_keys():
    doc = {"time": 0, "uid": new_uid()}
    schema_validators[DocumentNames.start].validate(doc)

    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.start].validate(doc)
    # Now add illegal key.
    doc.update({"b.": "c"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.start].validate(doc)

    doc = {
        "time": 0,
        "uid": new_uid(),
        "data_keys": {"a": {"source": "", "dtype": "number", "shape": []}},
        "run_start": new_uid(),
    }
    schema_validators[DocumentNames.descriptor].validate(doc)
    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.descriptor].validate(doc)
    # Now add illegal key.
    doc.update({"b.c": "d"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.descriptor].validate(doc)

    doc = {
        "time": 0,
        "uid": new_uid(),
        "exit_status": "success",
        "reason": "",
        "run_start": new_uid(),
    }
    schema_validators[DocumentNames.stop].validate(doc)
    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.stop].validate(doc)
    # Now add illegal key.
    doc.update({".b": "c"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.stop].validate(doc)



import pytest
from jsonschema.exceptions import ValidationError

import event_model


@pytest.fixture
def start():
    run_bundle = event_model.compose_run()
    return run_bundle.start_doc


def test_projection_in_start_doc():
    run_bundle = event_model.compose_run(
        uid="42", metadata={"projections": valid_projections}
    )
    start_doc = run_bundle.start_doc
    assert start_doc["projections"] == valid_projections


def test_projection_schema(start):
    start["projections"] = valid_projections
    event_model.schema_validators[event_model.DocumentNames.start].validate(start)


def test_bad_calc_field(start):
    bad_calc_projections = [
        # calc requires the calc fields
        {
            "name": "test",
            "version": "42.0.0",
            "configuration": {},
            "projection": {
                "linked_field": {
                    "type": "calculated",
                    "location": "event",
                    "stream": "primary",
                    "field": "ccd",
                },
            },
        },
    ]

    start["projections"] = bad_calc_projections
    with pytest.raises(
        ValidationError,
    ):
        event_model.schema_validators[event_model.DocumentNames.start].validate(start)


def test_bad_configuration_field(start):
    bad_configuration_projections = [
        {
            "name": "test",
            "version": "42.0.0",
            "configuration": {},
            "projection": {
                "bad_config_field": {
                    "type": "calculated",
                    "location": "event",
                    "config_index": 0,
                    "config_device": "camera",
                    "stream": "primary",
                    # "field": "setting"
                },
            },
        },
    ]

    start["projections"] = bad_configuration_projections
    with pytest.raises(
        ValidationError,
    ):
        event_model.schema_validators[event_model.DocumentNames.start].validate(start)


def test_bad_event_field(start):
    bad_event_projections = [
        {
            "name": "test",
            "version": "42.0.0",
            "configuration": {},
            "projection": {
                "bad_event_field": {
                    "type": "linked",
                    "location": "event",
                    # "stream": "primary",
                    "field": "ccd",
                },
            },
        },
    ]
    start["projections"] = bad_event_projections
    with pytest.raises(
        ValidationError,
    ):
        event_model.schema_validators[event_model.DocumentNames.start].validate(start)


def test_bad_location_field(start):
    bad_event_projections = [
        {
            "name": "test",
            "version": "42.0.0",
            "configuration": {},
            "projection": {
                "bad_event_field": {
                    "type": "linked",
                    "location": "event",
                    # "stream": "primary",
                    "field": "ccd",
                },
            },
        },
    ]
    start["projections"] = bad_event_projections
    with pytest.raises(
        ValidationError,
    ):
        event_model.schema_validators[event_model.DocumentNames.start].validate(start)


def test_bad_static_field(start):
    bad_event_projections = [
        {
            "name": "test",
            "version": "42.0.0",
            "configuration": {},
            "projection": {
                "bad_event_field": {
                    "type": "static",
                    "location": "event",
                    # "stream": "primary",
                    "field": "ccd",
                },
            },
        },
    ]
    start["projections"] = bad_event_projections
    with pytest.raises(
        ValidationError,
    ):
        event_model.schema_validators[event_model.DocumentNames.start].validate(start)


valid_projections = [
    {
        "name": "test",
        "version": "42.0.0",
        "configuration": {},
        "projection": {
            "linked_field": {
                "type": "linked",
                "location": "event",
                "stream": "primary",
                "field": "ccd",
            },
            "calc_field": {
                "type": "calculated",
                "location": "event",
                "field": "calc_field",
                "stream": "calc_stream",
                "calculation": {
                    "callable": "pizza.order:slice",
                    "kwargs": {"toppings": "cheese"},
                },
            },
            "config_field": {
                "type": "linked",
                "location": "configuration",
                "config_index": 0,
                "config_device": "camera",
                "stream": "primary",
                "field": "setting",
            },
            "static_field": {"type": "static", "value": "strcvsdf"},
        },
    }
]



from collections import defaultdict

import numpy
import pytest

import event_model
from event_model.documents.stream_datum import StreamRange


def test_run_router(tmp_path):
    bundle = event_model.compose_run()
    docs = []
    start_doc, compose_descriptor, compose_resource, compose_stop = bundle
    docs.append(("start", start_doc))
    bundle = compose_descriptor(
        data_keys={
            "motor": {"shape": [], "dtype": "number", "source": "..."},
            "image": {
                "shape": [512, 512],
                "dtype": "number",
                "source": "...",
                "external": "FILESTORE:",
            },
        },
        name="primary",
    )
    primary_descriptor_doc, compose_primary_event, compose_event_page = bundle
    docs.append(("descriptor", primary_descriptor_doc))
    bundle = compose_descriptor(
        data_keys={"motor": {"shape": [], "dtype": "number", "source": "..."}},
        name="baseline",
    )
    baseline_descriptor_doc, compose_baseline_event, compose_event_page = bundle
    docs.append(("descriptor", baseline_descriptor_doc))
    bundle = compose_resource(
        spec="TIFF", root=str(tmp_path), resource_path="stack.tiff", resource_kwargs={}
    )
    resource_doc, compose_datum, compose_datum_page = bundle
    docs.append(("resource", resource_doc))
    datum_doc = compose_datum(datum_kwargs={"slice": 5})
    docs.append(("datum", datum_doc))
    primary_event_doc = compose_primary_event(
        data={"motor": 0, "image": datum_doc["datum_id"]},
        timestamps={"motor": 0, "image": 0},
        filled={"image": False},
    )
    docs.append(("event", primary_event_doc))
    baseline_event_doc = compose_baseline_event(
        data={"motor": 0}, timestamps={"motor": 0}
    )
    docs.append(("event", baseline_event_doc))
    stop_doc = compose_stop()
    docs.append(("stop", stop_doc))

    # Empty list of factories. Just make sure nothing blows up.
    rr = event_model.RunRouter([])
    for name, doc in docs:
        rr(name, doc)

    # A factory that rejects all runs.
    def null_factory(name, doc):
        return [], []

    rr = event_model.RunRouter([null_factory])
    for name, doc in docs:
        rr(name, doc)

    # A factory that accepts all runs.
    collected = []

    def collector(name, doc):
        if name == "event_page":
            name = "event"
            (doc,) = event_model.unpack_event_page(doc)
        elif name == "datum_page":
            name = "datum"
            (doc,) = event_model.unpack_datum_page(doc)
        collected.append((name, doc))

    def all_factory(name, doc):
        return [collector], []

    rr = event_model.RunRouter([all_factory])
    for name, doc in docs:
        rr(name, doc)

    assert collected == docs
    collected.clear()

    # A factory that returns a subfactory interested in 'baseline' only.
    def subfactory(name, doc):
        if doc.get("name") == "baseline":
            return [collector]
        return []

    def factory_with_subfactory_only(name, doc):
        return [], [subfactory]

    rr = event_model.RunRouter([factory_with_subfactory_only])
    for name, doc in docs:
        rr(name, doc)

    expected_item = ("event", baseline_event_doc)
    unexpected_item = ("event", primary_event_doc)
    assert expected_item in collected
    assert unexpected_item not in collected
    collected.clear()

    # Test factory that expects old (pre-1.14.0) RunRouter behavior.

    collected_header_docs = {}

    class LocalException3(Exception):
        ...

    def header_collector(name, doc):
        if name in ("start", "stop", "descriptor"):
            key = (name, doc["uid"])
            if key in collected_header_docs:
                raise LocalException3
            collected_header_docs[key] = doc

    def all_factory(name, doc):
        header_collector(name, doc)
        return [header_collector], []

    rr = event_model.RunRouter([all_factory])
    with pytest.warns(UserWarning, match="1.14.0"), pytest.raises(LocalException3):
        for name, doc in docs:
            rr(name, doc)

    collected_header_docs.clear()

    # Test subfactory that expects old (pre-1.14.0) RunRouter behavior.

    def factory_with_subfactory_only(name, doc):
        header_collector(name, doc)

        def subfactory(name, doc):
            if doc.get("name") == "baseline":
                header_collector(name, doc)
                return [header_collector]
            return []

        return [], [subfactory]

    rr = event_model.RunRouter([factory_with_subfactory_only])
    with pytest.warns(UserWarning, match="1.14.0"), pytest.raises(LocalException3):
        for name, doc in docs:
            rr(name, doc)

    collected_header_docs.clear()

    # Test RunRouter with handler_registry.

    class FakeTiffHandler:
        def __init__(self, resource_path):
            assert resource_path == str(tmp_path / "stack.tiff")

        def __call__(self, slice):
            return numpy.ones((5, 5))

    reg = {"TIFF": FakeTiffHandler}

    def check_filled(name, doc):
        if name == "event_page":
            for is_filled in doc["filled"].values():
                assert all(is_filled)
        elif name == "event":
            for is_filled in doc["filled"].values():
                assert is_filled

    def check_not_filled(name, doc):
        if name == "event_page":
            for is_filled in doc["filled"].values():
                assert not any(is_filled)
        elif name == "event":
            for is_filled in doc["filled"].values():
                assert not is_filled

    def check_filled_factory(name, doc):
        return [check_filled], []

    def check_not_filled_factory(name, doc):
        return [check_not_filled], []

    # If reg is missing our spec (or just not given) docs pass through
    # unfilled.
    rr = event_model.RunRouter([check_not_filled_factory])
    for name, doc in docs:
        rr(name, doc)

    # If fill_or_fail is set to True and reg is missing our spec (or just not
    # given) we raise.
    rr = event_model.RunRouter([check_not_filled_factory], fill_or_fail=True)
    with pytest.raises(event_model.UndefinedAssetSpecification):
        for name, doc in docs:
            rr(name, doc)

    # If spec is provided, docs are filled, regardless of fill_or_fail.
    rr = event_model.RunRouter([check_filled_factory], reg)
    for name, doc in docs:
        rr(name, doc)

    rr = event_model.RunRouter([check_filled_factory], reg, fill_or_fail=True)
    for name, doc in docs:
        rr(name, doc)


def test_run_router_streams(tmp_path):
    bundle = event_model.compose_run()
    docs = []
    start_doc, compose_stream_resource, stop_doc = (
        bundle.start_doc,
        bundle.compose_stream_resource,
        bundle.compose_stop(),
    )
    docs.append(("start", start_doc))
    stream_resource_doc, compose_stream_datum = compose_stream_resource(
        spec="TIFF_STREAM",
        data_key="det1",
        root=str(tmp_path),
        resource_path="test_streams",
        resource_kwargs={},
    )
    docs.append(("stream_resource", stream_resource_doc))
    datum_doc = compose_stream_datum(
        StreamRange(start=0, stop=0), StreamRange(start=0, stop=0)
    )
    docs.append(("stream_datum", datum_doc))
    docs.append(("stop", stop_doc))

    # Empty list of factories. Just make sure nothing blows up.
    rr = event_model.RunRouter([])
    for name, doc in docs:
        rr(name, doc)

    # Collector factory for stream resource and data
    resource_list = []
    data_list = []

    def collector(name, doc):
        if name == "stream_resource":
            resource_list.append((name, doc))
        elif name == "stream_datum":
            data_list.append((name, doc))

    def all_factory(name, doc):
        return [collector], []

    rr = event_model.RunRouter([all_factory])
    for name, doc in docs:
        rr(name, doc)
    assert len(resource_list) == 1
    assert len(data_list) == 1


def test_subfactory():
    # this test targeted the bug described in issue #170
    factory_documents = defaultdict(list)
    subfactory_documents = defaultdict(list)

    def factory(name, start_doc):
        def collect_factory_documents(name, doc):
            factory_documents[name].append(doc)

        def collect_subfactory_documents(name, doc):
            subfactory_documents[name].append(doc)

        def subfactory(name, descriptor_doc):
            return [collect_subfactory_documents]

        return [collect_factory_documents], [subfactory]

    rr = event_model.RunRouter([factory])

    run_bundle = event_model.compose_run()
    rr("start", run_bundle.start_doc)
    assert len(factory_documents) == 1
    assert len(factory_documents["start"]) == 1
    assert factory_documents["start"] == [run_bundle.start_doc]
    assert len(subfactory_documents) == 0

    descriptor_bundle = run_bundle.compose_descriptor(
        data_keys={"motor": {"shape": [], "dtype": "number", "source": "..."}},
        name="primary",
    )
    rr("descriptor", descriptor_bundle.descriptor_doc)
    assert len(factory_documents) == 2
    assert len(factory_documents["start"]) == 1
    assert factory_documents["start"] == [run_bundle.start_doc]
    assert len(factory_documents["descriptor"]) == 1
    assert factory_documents["descriptor"] == [descriptor_bundle.descriptor_doc]

    assert len(subfactory_documents) == 2
    assert len(subfactory_documents["start"]) == 1
    assert subfactory_documents["start"] == [run_bundle.start_doc]
    assert len(subfactory_documents["descriptor"]) == 1
    assert subfactory_documents["descriptor"] == [descriptor_bundle.descriptor_doc]

    stop_doc = run_bundle.compose_stop()
    rr("stop", stop_doc)

    assert len(rr._start_to_start_doc) == 0


def test_same_start_doc_twice():
    "If the user sends us the same uid twice, raise helpfully."
    rr = event_model.RunRouter([])
    doc = {"time": 0, "uid": "stuff"}
    rr("start", doc)
    with pytest.raises(ValueError):
        rr("start", doc)  # exact same object
    with pytest.raises(ValueError):
        rr("start", doc.copy())  # same content
    doc2 = {"time": 1, "uid": "stuff"}
    with pytest.raises(ValueError):
        rr("start", doc2)  # same uid, different content


def test_subfactory_callback_exception():
    """
    Test that RunRouter._start_to_descriptors and RunRouter._descriptor_to_start
    are updated in the case that a subfactory callback raises an Exception.
    """

    def exception_callback(name, doc):
        """a callback that always raises Exception"""
        raise Exception()

    def exception_callback_subfactory(descriptor_doc_name, descriptor_doc):
        """a subfactory that always returns one exception_callback"""
        return [exception_callback]

    def exception_callback_factory(start_doc_name, start_doc):
        """a factory that return 0 callbacks and one exception_callback_subfactory"""
        return ([], [exception_callback_subfactory])

    rr = event_model.RunRouter([exception_callback_factory])

    start_doc = {"time": 0, "uid": "abcdef"}
    rr("start", start_doc)

    descriptor_doc = {"run_start": "abcdef", "uid": "ghijkl"}
    with pytest.raises(Exception):
        rr("descriptor", descriptor_doc)

    assert rr._start_to_descriptors["abcdef"] == ["ghijkl"]
    assert rr._descriptor_to_start["ghijkl"] == "abcdef"

    event_document = {"descriptor": "ghijkl", "uid": "mnopqr"}
    rr.event(event_document)



import json
import logging
from argparse import ArgumentParser
from pathlib import Path
from subprocess import CalledProcessError, check_output
from typing import List, Optional


def report_output(stdout: bytes, label: str) -> List[str]:
    ret = stdout.decode().strip().split("\n")
    print(f"{label}: {ret}")
    return ret


def get_branch_contents(ref: str) -> List[str]:
    """Get the list of directories in a branch."""
    stdout = check_output(["git", "ls-tree", "-d", "--name-only", ref])
    return report_output(stdout, "Branch contents")


def get_sorted_tags_list() -> List[str]:
    """Get a list of sorted tags in descending order from the repository."""
    stdout = check_output(["git", "tag", "-l", "--sort=-v:refname"])
    return report_output(stdout, "Tags list")


def get_versions(ref: str, add: Optional[str], remove: Optional[str]) -> List[str]:
    """Generate the file containing the list of all GitHub Pages builds."""
    # Get the directories (i.e. builds) from the GitHub Pages branch
    try:
        builds = set(get_branch_contents(ref))
    except CalledProcessError:
        builds = set()
        logging.warning(f"Cannot get {ref} contents")

    # Add and remove from the list of builds
    if add:
        builds.add(add)
    if remove:
        assert remove in builds, f"Build '{remove}' not in {sorted(builds)}"
        builds.remove(remove)

    # Get a sorted list of tags
    tags = get_sorted_tags_list()

    # Make the sorted versions list from main branches and tags
    versions: List[str] = []
    for version in ["master", "main"] + tags:
        if version in builds:
            versions.append(version)
            builds.remove(version)

    # Add in anything that is left to the bottom
    versions += sorted(builds)
    print(f"Sorted versions: {versions}")
    return versions


def write_json(path: Path, repository: str, versions: str):
    org, repo_name = repository.split("/")
    struct = [
        dict(version=version, url=f"https://{org}.github.io/{repo_name}/{version}/")
        for version in versions
    ]
    text = json.dumps(struct, indent=2)
    print(f"JSON switcher:\n{text}")
    path.write_text(text)


def main(args=None):
    parser = ArgumentParser(
        description="Make a versions.txt file from gh-pages directories"
    )
    parser.add_argument(
        "--add",
        help="Add this directory to the list of existing directories",
    )
    parser.add_argument(
        "--remove",
        help="Remove this directory from the list of existing directories",
    )
    parser.add_argument(
        "repository",
        help="The GitHub org and repository name: ORG/REPO",
    )
    parser.add_argument(
        "output",
        type=Path,
        help="Path of write switcher.json to",
    )
    args = parser.parse_args(args)

    # Write the versions file
    versions = get_versions("origin/gh-pages", args.add, args.remove)
    write_json(args.output, args.repository, versions)


if __name__ == "__main__":
    main()



# type: ignore

import sys

import pytest

if sys.version_info[:2] >= (3, 9):
    # Test schema generation
    import json
    import os

    import event_model
    from event_model.documents import (
        Datum,
        DatumPage,
        Event,
        EventDescriptor,
        EventPage,
        Resource,
        RunStart,
        RunStop,
        StreamDatum,
        StreamResource,
    )
    from event_model.documents.generate.typeddict_to_schema import (
        parse_typeddict_to_schema,
    )

    typed_dict_class_list = [
        DatumPage,
        Datum,
        EventDescriptor,
        EventPage,
        Event,
        Resource,
        RunStart,
        RunStop,
        StreamDatum,
        StreamResource,
    ]

    SCHEMA_PATH = event_model.__path__[0] + "/schemas/"

    @pytest.mark.parametrize("typed_dict_class", typed_dict_class_list)
    def test_generated_json_matches_typed_dict(typed_dict_class, tmpdir):
        parse_typeddict_to_schema(typed_dict_class, out_dir=tmpdir)
        file_name = os.listdir(tmpdir)[0]
        generated_file_path = os.path.join(tmpdir, file_name)
        old_file_path = os.path.join(SCHEMA_PATH, file_name)

        with open(generated_file_path) as generated_file, open(
            old_file_path
        ) as old_file:
            try:
                assert json.load(generated_file) == json.load(old_file)
            except AssertionError:
                raise Exception(
                    f"`{typed_dict_class.__name__}` can generate a json schema, but "
                    f"it doesn't match the schema in `{SCHEMA_PATH}`. Did you forget "
                    "to run `python event_model/typeddict_to_schema.py` after changes "
                    f"to `{typed_dict_class.__name__}`?"
                )

else:
    # Test an error is thrown for pyton <= 3.8
    def test_schema_generation_import_throws_error():
        with pytest.raises(EnvironmentError):
            from event_model.documents.generate import typeddict_to_schema

            typeddict_to_schema



from typing import Any, Dict

from typing_extensions import Annotated, Literal, NotRequired, TypedDict

from .generate.type_wrapper import AsRef, Field, add_extra_schema

RUN_STOP_EXTRA_SCHEMA = {
    "patternProperties": {"^([^./]+)$": {"$ref": "#/definitions/DataType"}},
    "additionalProperties": False,
}


@add_extra_schema(RUN_STOP_EXTRA_SCHEMA)
class RunStop(TypedDict):
    """
    Document for the end of a run indicating the success/fail state of the
    run and the end time
    """

    data_type: NotRequired[
        Annotated[Any, Field(description="data_type"), AsRef("DataType")]
    ]
    exit_status: Annotated[
        Literal["success", "abort", "fail"],
        Field(description="State of the run when it ended"),
    ]
    num_events: NotRequired[
        Annotated[
            Dict[str, int],
            Field(
                description="Number of Events per named stream",
            ),
        ]
    ]
    reason: NotRequired[
        Annotated[str, Field(description="Long-form description of why the run ended")]
    ]
    run_start: Annotated[
        str,
        Field(
            description="Reference back to the run_start document that this document "
            "is paired with.",
        ),
    ]
    time: Annotated[float, Field(description="The time the run ended. Unix epoch")]
    uid: Annotated[str, Field(description="Globally unique ID for this document")]



from typing import Any, Dict, List, Union

from typing_extensions import Annotated, NotRequired, TypedDict

from .generate.type_wrapper import AsRef, Field, add_extra_schema

EVENT_PAGE_EXTRA_SCHEMA = {"additionalProperties": False}


class PartialEventPage(TypedDict):
    data: Annotated[
        Dict[str, List[Any]],
        AsRef("Dataframe"),
        Field(description="The actual measurement data"),
    ]
    filled: NotRequired[
        Annotated[
            Dict[str, List[Union[bool, str]]],
            AsRef("DataframeForFilled"),
            Field(
                description="Mapping each of the keys of externally-stored data to an "
                "array containing the boolean False, indicating that the data has not "
                "been loaded, or to foreign keys (moved here from 'data' when the data "
                "was loaded)"
            ),
        ]
    ]
    timestamps: Annotated[
        Dict[str, List[Any]],
        AsRef("Dataframe"),
        Field(description="The timestamps of the individual measurement data"),
    ]
    time: Annotated[
        List[float],
        Field(
            description="Array of Event times. This maybe different than the "
            "timestamps on each of the data entries"
        ),
    ]


@add_extra_schema(EVENT_PAGE_EXTRA_SCHEMA)
class EventPage(PartialEventPage):
    """Page of documents to record a quanta of collected data"""

    descriptor: Annotated[
        str,
        Field(
            description="The UID of the EventDescriptor to which all of the Events in "
            "this page belong",
        ),
    ]

    seq_num: Annotated[
        List[int],
        Field(
            description="Array of sequence numbers to identify the location of each "
            "Event in the Event stream",
        ),
    ]
    uid: Annotated[
        List[str],
        Field(description="Array of globally unique identifiers for each Event"),
    ]



from typing import Any, Dict

from typing_extensions import Annotated, Literal, NotRequired, TypedDict

from .generate.type_wrapper import Field, add_extra_schema

STREAM_RESOURCE_EXTRA_SCHEMA = {"additionalProperties": False}


@add_extra_schema(STREAM_RESOURCE_EXTRA_SCHEMA)
class StreamResource(TypedDict):
    """
    Document to reference a collection (e.g. file or group of files) of
    externally-stored data streams
    """

    path_semantics: NotRequired[
        Annotated[
            Literal["posix", "windows"],
            Field(description="Rules for joining paths"),
        ]
    ]
    data_key: Annotated[
        str,
        Field(
            description="A string to show which data_key of the "
            "Descriptor are being streamed"
        ),
    ]
    resource_kwargs: Annotated[
        Dict[str, Any],
        Field(
            description="Additional argument to pass to the Handler to read a "
            "Stream Resource",
        ),
    ]
    resource_path: Annotated[
        str, Field(description="Filepath or URI for locating this resource")
    ]
    root: Annotated[
        str,
        Field(
            description="Subset of resource_path that is a local detail, not semantic."
        ),
    ]
    run_start: NotRequired[
        Annotated[
            str,
            Field(
                description="Globally unique ID to the run_start document "
                "this Stream Resource is associated with.",
            ),
        ]
    ]
    spec: Annotated[
        str,
        Field(
            description="String identifying the format/type of this Stream Resource, "
            "used to identify a compatible Handler",
        ),
    ]
    uid: Annotated[
        str, Field(description="Globally unique identifier for this Stream Resource")
    ]



from typing import Any, Dict, List, Union

from typing_extensions import Annotated, Literal, NotRequired, TypedDict

from event_model.documents.generate.type_wrapper import Field, add_extra_schema

from .generate.type_wrapper import AsRef


class Hints(TypedDict):
    """Start-level hints"""

    dimensions: NotRequired[
        Annotated[
            List[List[Union[List[str], str]]],
            Field(
                description="The independent axes of the experiment. "
                "Ordered slow to fast",
            ),
        ]
    ]


class Calculation(TypedDict):
    args: NotRequired[List]
    callable: Annotated[
        str, Field(description="callable function to perform calculation")
    ]
    kwargs: NotRequired[
        Annotated[Dict[str, Any], Field(description="kwargs for calcalation callable")]
    ]


class Projection(TypedDict):
    """Where to get the data from"""

    calculation: NotRequired[
        Annotated[
            Calculation,
            Field(
                description="required fields if type is calculated",
                title="calculation properties",
            ),
        ]
    ]
    config_index: NotRequired[int]
    config_device: NotRequired[str]
    field: NotRequired[str]
    location: NotRequired[
        Annotated[
            Literal["start", "event", "configuration"],
            Field(
                description="start comes from metadata fields in the start document, "
                "event comes from event, configuration comes from configuration "
                "fields in the event_descriptor document"
            ),
        ]
    ]
    stream: NotRequired[str]
    type: NotRequired[
        Annotated[
            Literal["linked", "calculated", "static"],
            Field(
                description="linked: a value linked from the data set, "
                "calculated: a value that requires calculation, "
                "static:  a value defined here in the projection ",
            ),
        ]
    ]
    value: NotRequired[
        Annotated[
            Any,
            Field(
                description="value explicitely defined in the projection "
                "when type==static."
            ),
        ]
    ]


RUN_START_EXTRA_SCHEMA = {
    "definitions": {
        "DataType": {
            "patternProperties": {"^([^./]+)$": {"$ref": "#/definitions/DataType"}},
            "additionalProperties": False,
        },
        "Projection": {
            "allOf": [
                {
                    "if": {
                        "allOf": [
                            {"properties": {"location": {"enum": ["configuration"]}}},
                            {"properties": {"type": {"enum": ["linked"]}}},
                        ]
                    },
                    "then": {
                        "required": [
                            "type",
                            "location",
                            "config_index",
                            "config_device",
                            "field",
                            "stream",
                        ]
                    },
                },
                {
                    "if": {
                        "allOf": [
                            {"properties": {"location": {"enum": ["event"]}}},
                            {"properties": {"type": {"enum": ["linked"]}}},
                        ]
                    },
                    "then": {"required": ["type", "location", "field", "stream"]},
                },
                {
                    "if": {
                        "allOf": [
                            {"properties": {"location": {"enum": ["event"]}}},
                            {"properties": {"type": {"enum": ["calculated"]}}},
                        ]
                    },
                    "then": {"required": ["type", "field", "stream", "calculation"]},
                },
                {
                    "if": {"properties": {"type": {"enum": ["static"]}}},
                    "then": {"required": ["type", "value"]},
                },
            ],
        },
    },
    "properties": {
        "hints": {
            "additionalProperties": False,
            "patternProperties": {"^([^.]+)$": {"$ref": "#/definitions/DataType"}},
        },
    },
    "patternProperties": {"^([^./]+)$": {"$ref": "#/definitions/DataType"}},
    "additionalProperties": False,
}


class Projections(TypedDict):
    """Describe how to interperet this run as the given projection"""

    configuration: Annotated[
        Dict[str, Any], Field(description="Static information about projection")
    ]
    name: NotRequired[Annotated[str, Field(description="The name of the projection")]]
    projection: Annotated[Dict[Any, Projection], Field(description="")]
    version: Annotated[
        str,
        Field(
            description="The version of the projection spec. Can specify the version "
            "of an external specification.",
        ),
    ]


@add_extra_schema(RUN_START_EXTRA_SCHEMA)
class RunStart(TypedDict):
    """
    Document created at the start of run. Provides a seach target and
    later documents link to it
    """

    data_groups: NotRequired[
        Annotated[
            List[str],
            Field(
                description="An optional list of data access groups that have meaning "
                "to some external system. Examples might include facility, beamline, "
                "end stations, proposal, safety form.",
            ),
        ]
    ]
    data_session: NotRequired[
        Annotated[
            str,
            Field(
                description="An optional field for grouping runs. The meaning is "
                "not mandated, but this is a data management grouping and not a "
                "scientific grouping. It is intended to group runs in a visit or "
                "set of trials.",
            ),
        ]
    ]
    data_type: NotRequired[Annotated[Any, Field(description=""), AsRef("DataType")]]
    group: NotRequired[
        Annotated[str, Field(description="Unix group to associate this data with")]
    ]
    hints: NotRequired[Annotated[Hints, Field(description="Start-level hints")]]
    owner: NotRequired[
        Annotated[str, Field(description="Unix owner to associate this data with")]
    ]
    project: NotRequired[
        Annotated[str, Field(description="Name of project that this run is part of")]
    ]
    projections: NotRequired[Annotated[List[Projections], Field(description="")]]
    sample: NotRequired[
        Annotated[
            Union[Dict[str, Any], str],
            Field(
                description="Information about the sample, may be a UID to "
                "another collection"
            ),
        ]
    ]
    scan_id: NotRequired[
        Annotated[int, Field(description="Scan ID number, not globally unique")]
    ]
    time: Annotated[float, Field(description="Time the run started.  Unix epoch time")]
    uid: Annotated[str, Field(description="Globally unique ID for this run")]



from typing import Any, Dict, Union

from typing_extensions import Annotated, NotRequired, TypedDict

from .generate.type_wrapper import Field, add_extra_schema

EVENT_EXTRA_SCHEMA = {"additionalProperties": False}


class PartialEvent(TypedDict):
    data: Annotated[Dict[str, Any], Field(description="The actual measurement data")]
    filled: NotRequired[
        Annotated[
            Dict[str, Union[bool, str]],
            Field(
                description="Mapping each of the keys of externally-stored data to the "
                "boolean False, indicating that the data has not been loaded, or to "
                "foreign keys (moved here from 'data' when the data was loaded)"
            ),
        ]
    ]
    time: Annotated[
        float,
        Field(
            description="The event time. This maybe different than the timestamps on "
            "each of the data entries.",
        ),
    ]
    timestamps: Annotated[
        Dict[str, Any],
        Field(description="The timestamps of the individual measurement data"),
    ]


@add_extra_schema(EVENT_EXTRA_SCHEMA)
class Event(PartialEvent):
    """Document to record a quanta of collected data"""

    descriptor: Annotated[
        str, Field(description="UID of the EventDescriptor to which this Event belongs")
    ]
    seq_num: Annotated[
        int,
        Field(
            description="Sequence number to identify the location of this Event in the "
            "Event stream",
        ),
    ]
    uid: Annotated[str, Field(description="Globally unique identifier for this Event")]



"""
A wrapper used to patch out schema generation utilities.
"""


from dataclasses import dataclass

pydantic_version = None
try:
    try:
        from pydantic import v1 as pydantic  # type: ignore
    except ImportError:
        import pydantic

    pydantic_version = pydantic.__version__

    Field = pydantic.Field
    FieldInfo = pydantic.fields.FieldInfo
    BaseConfig = pydantic.BaseConfig
    BaseModel = pydantic.BaseModel
    create_model = pydantic.create_model
except ModuleNotFoundError:

    def Field(*args, **kwargs):  # type: ignore
        ...

    class FieldInfo:  # type: ignore
        ...

    class BaseConfig:  # type: ignore
        ...

    class BaseModel:  # type: ignore
        ...

    def create_model(*args, **kwargs):  # type: ignore
        ...


extra_schema = {}


def add_extra_schema(schema: dict):
    def inner(cls):
        extra_schema[cls] = schema
        return cls

    return inner


@dataclass
class AsRef:
    ref_name: str


# We need to check that only one element in an annotation is the type of
# the field, the others have to be instances of classes in this tuple
ALLOWED_ANNOTATION_ELEMENTS = (AsRef, FieldInfo)



from typing import Any, Dict

from typing_extensions import Annotated, Literal, NotRequired, TypedDict

from .generate.type_wrapper import Field, add_extra_schema

RESOURCE_EXTRA_SCHEMA = {"additionalProperties": False}


class PartialResource(TypedDict):
    spec: Annotated[
        str,
        Field(
            description="String identifying the format/type of this Resource, used to "
            "identify a compatible Handler"
        ),
    ]
    resource_path: Annotated[
        str, Field(description="Filepath or URI for locating this resource")
    ]
    resource_kwargs: Annotated[
        Dict[str, Any],
        Field(
            description="Additional argument to pass to the Handler to read a Resource"
        ),
    ]

    root: Annotated[
        str,
        Field(
            description="Subset of resource_path that is a local detail, not semantic."
        ),
    ]
    uid: Annotated[
        str, Field(description="Globally unique identifier for this Resource")
    ]


@add_extra_schema(RESOURCE_EXTRA_SCHEMA)
class Resource(PartialResource):
    """
    Document to reference a collection (e.g. file or group of files) of
    externally-stored data
    """

    path_semantics: NotRequired[
        Annotated[
            Literal["posix", "windows"],
            Field(description="Rules for joining paths"),
        ]
    ]
    run_start: NotRequired[
        Annotated[
            str,
            Field(
                description="Globally unique ID to the run_start document this "
                "resource is associated with.",
            ),
        ]
    ]



# flake8: noqa
from .datum import Datum
from .datum_page import DatumPage
from .event import Event
from .event_descriptor import EventDescriptor
from .event_page import EventPage
from .resource import Resource
from .run_start import RunStart
from .run_stop import RunStop
from .stream_datum import StreamDatum
from .stream_resource import StreamResource



# type: ignore

import json
import re
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Optional, Tuple, Type, Union

from typing_extensions import (
    Annotated,
    NotRequired,
    _TypedDictMeta,
    get_args,
    get_origin,
)

from event_model import SCHEMA_PATH
from event_model.documents import (
    Datum,
    DatumPage,
    Event,
    EventDescriptor,
    EventPage,
    Resource,
    RunStart,
    RunStop,
    StreamDatum,
    StreamResource,
)
from event_model.documents.generate.type_wrapper import (
    ALLOWED_ANNOTATION_ELEMENTS,
    AsRef,
    BaseConfig,
    BaseModel,
    Field,
    FieldInfo,
    create_model,
    extra_schema,
    pydantic_version,
)

# The hacky indexing on types isn't possible with python < 3.9
if sys.version_info[:2] < (3, 9):
    raise EnvironmentError("schema generation requires python 3.8 or higher")

if not pydantic_version:
    raise EnvironmentError(
        "schema generation requires pydantic < 2.0 to run, pydantic isn't installed"
    )


SCHEMA_OUT_DIR = Path("event_model") / SCHEMA_PATH


# Used to add user written schema to autogenerated schema.
def merge_dicts(dict1: dict, dict2: dict) -> dict:
    """
    Takes two dictionaries with subdirectories and returns a new dictionary of
    the two merged:
    dict1 = {
        "x1": {
            "y1": 0,  "y3": {"z1" : [1, 2], "z2": 1}
        },
        "x2" : 0,
        "x3": 1
    }
    and
    dict2 = {
        "x1": {
            "y2" : 0,  "y3": {"z1": [3, 4], "z3": 5}
        },
        "x3" : 0
    }
    returns
    {
        "x1": {
            "y1": 0, "y2": 0,  "y3": {"z1": [1, 2, 3, 4], "z2": 1, "z3": 5}
        },
        "x2": 0
        "x3": 1
    }
    """

    return_dict = dict2.copy()

    for key in dict1:
        if key not in dict2:
            return_dict[key] = dict1[key]

        elif not isinstance(dict1[key], type(dict2[key])):
            return_dict[key] = dict1[key]

        elif isinstance(dict1[key], dict):
            return_dict[key] = merge_dicts(dict1[key], dict2[key])

        elif isinstance(dict1[key], list):
            return_dict[key] = dict1[key] + dict2[key]

    return return_dict


def format_all_annotations(
    typed_dict: _TypedDictMeta, new_basemodel_classes: Dict[str, BaseModel] = None
) -> _TypedDictMeta:
    """Goes through all field type annotations and formats them to be acceptable
    to pydantic.

    It formats annotations in the following way:
    * NotRequired[Annotated[X, FieldInfo, ...] -> Annotated[Optional[X],
      FieldInfo, ...]
    * Annotated[X, AsRef("ref_name"), FieldInfo] ->
        Annotated[ref_name_BaseModel_containing_X, FieldInfo]
    * If X is also a typeddict then Annotated[X, FieldInfo] ->
        Annotated[parse_typeddict_to_schema(X), FieldInfo]

    Parameters
    ----------
    typed_dict : dict
        TypedDict to change annotations of prior to being converted to BaseModel.

    new_basemodel_classes : Optional[Dict[str, BaseModel]
        Dict where keys are names of _TypedDictMeta classes referenced inside
        the TypedDict being converted to a BaseModel, and values are BaseModel
        classes that have already been generated from those TypedDicts -
        pydantic will throw an error if we try to generate the same BaseModel
        twice so we just reference the one already created if multiple
        annotations have the same TypedDict.

    Returns
    -------
    typed_dict : dict
        The same dict with field annotations adjusted to be pydantic friendly.

    """
    new_annotations = {}

    if new_basemodel_classes is None:
        new_basemodel_classes = {}

    for field_name, field_type in typed_dict.__annotations__.items():
        new_annotations[field_name] = field_parser(
            field_name, field_type, new_basemodel_classes
        )

    typed_dict.__annotations__ = new_annotations

    return typed_dict


def field_is_annotated(field: type) -> bool:
    """Returns True if a field is of type Annotated[X]."""
    return get_origin(field) == Annotated


def field_is_not_required(
    field_type: type, remove_origin_if_NotRequired: bool = False
) -> Tuple[type, bool]:
    """Checks if a field is of type is NotRequired[X].

    Parameters
    ----------
    field_type : dict
        Field type taken from an annotation.
    remove_origin_if_NotRequired : bool
        If True returns the inputted field_type with NotRequired stripped off.

    Returns
    -------
    (field_type, is_not_required) : Tuple[type, bool]
        is_not_required is True if the field is of type NotRequired[X].
        field_type is the same as the inputted field_type however if
        remove_origin_if_NotRequired is true it will strip NotRequired.
    """
    is_not_required = get_origin(field_type) == NotRequired
    if is_not_required and remove_origin_if_NotRequired:
        args = get_args(field_type)
        assert len(args) == 1
        field_type = args[0]
    return field_type, is_not_required


def get_field_type(
    field_name: str,
    field_type: type,
    new_basemodel_classes: Dict[str, BaseModel],
) -> type:
    """Goes through an annotation and finds the type it represents without the
    additional context - e.g int inside Annotated[int, Field(description=""),
    AsRef("blah")].
    It is assumed that either the field origin is Annotated, or the field has no
    origin.

    Parameters
    ----------
    field_name : str
        Name of the field being parsed.
    field_type : type
        Annotation to be parsed.
    new_basemodel_classes : dict
        Dict where keys are names of _TypedDictMeta classes referenced inside
        the TypedDict being converted to a BaseModel, and values are BaseModel
        classes that have already been generated from those TypedDicts -
        pydantic will throw an error if we try to generate the same BaseModel
        twice so we just reference the one already created if multiple
        annotations have the same TypedDict.

    Returns
    -------
    field_type : type
        The data type inside the annotation seperate from the other context in
        the annotation.
    """
    args = get_args(field_type)
    if args:
        field_type = [
            x
            for x in args
            if True not in [isinstance(x, y) for y in ALLOWED_ANNOTATION_ELEMENTS]
        ]
        assert len(field_type) == 1, (
            f'Field "{field_name}" has multiple types: '
            f'{"and ".join([x.__class__.__name__ for x in field_type])}'
        )
        field_type = field_type[0]

    # If the TypedDict references another TypedDict then another
    # BaseModel is recursively generated from that TypedDict,
    # and the field annotation is swapped to that BaseModel.
    field_type = change_sub_typed_dicts_to_basemodels(field_type, new_basemodel_classes)

    return field_type


def get_annotation_contents(field_type: type) -> Tuple[Optional[AsRef], FieldInfo]:
    """Goes through the args of an Annotation and parses out any AsRef(), or
    FieldInfo.

    Parameters
    ----------
    field_type : type
        Annotation to be parsed.

    Returns
    -------
    (as_ref, field_info) : Tuple[AsRef, FieldInfo]
        as_ref is the AsRef tag in the annotation, or None if
        there is no AsRef tag. field_info is the FieldInfo class returned from
        the Field() call in the Annotation, or an empty Field() call if none is found.
    """

    args = get_args(field_type)
    as_ref = None
    field_info = Field()
    if args:
        for arg in args:
            if isinstance(arg, AsRef):
                as_ref = arg
            elif isinstance(arg, FieldInfo):
                field_info = arg

    return as_ref, field_info


def parse_AsRef(
    field_type: type,
    field_info: FieldInfo,
    as_ref: AsRef,
    new_basemodel_classes: Dict[str, BaseModel],
) -> type:
    """Parses the AsRef tag and makes a new BaseModel class containing a
    __root__ of the field with the AsRef.

    Parameters
    -----------------
    field_type : type
        Datatype to put in the __root__ field of the generated basemodel.
    field_info : FieldInfo
        Info included in the Field() of the annotation.
    as_ref : AsRef
        AsRef tag in the annotation.
    new_basemodel_classes : dict
        Dict where keys are names of a basemodel class generated by this
        function, and values are the class itseldf. If another field has the
        same AsRef the already generated baseclass in this dict is used as the
        new field value.

    Returns
    -----------------
    field_type : type
        The generated basemodel.
    """
    ref_field_info = Field()
    # Regex is extracted from the field info and placed in the new field info
    if field_info.regex:
        ref_field_info.regex = field_info.regex
    ref_field_type = Annotated[field_type, ref_field_info]

    if as_ref.ref_name not in new_basemodel_classes:
        field_type = create_model(
            as_ref.ref_name,
            __config__=Config,
            __root__=(ref_field_type, None),
        )
        new_basemodel_classes[as_ref.ref_name] = field_type
    else:
        generated_basemodel_type = new_basemodel_classes[
            as_ref.ref_name
        ].__annotations__["__root__"]
        assert get_args(ref_field_type)[0] == get_args(generated_basemodel_type)[0], (
            f'Fields with type AsRef("{as_ref.ref_name}") have differing types: '
            f"{generated_basemodel_type} and {ref_field_type}"
        )
        field_type = new_basemodel_classes[as_ref.ref_name]

    return field_type


def change_sub_typed_dicts_to_basemodels(
    field_type: type, new_basemodel_classes: Dict[str, BaseModel]
) -> dict:
    """Checks for any TypedDicts in the field_type and converts them to
    basemodels.

    Parameters
    -----------------
    field_type : type
        Annotation to be parsed.
    new_basemodel_classes : dict
        Dict where keys are names of _TypedDictMeta classes referenced inside
        the TypedDict being converted to a BaseModel, and values are BaseModel
        classes that have already been generated from those TypedDicts -
        pydantic will throw an error if we try to generate the same BaseModel
        twice so we just reference the one already created if multiple
        annotations have the same TypedDict.

    Returns
    -----------------
    field_type : type
        New field type with TypedDicts swapped to basemodels
    """

    if isinstance(field_type, _TypedDictMeta):
        if field_type.__name__ not in new_basemodel_classes:
            field_type = parse_typeddict_to_schema(
                field_type,
                return_basemodel=True,
                new_basemodel_classes=new_basemodel_classes,
            )
            new_basemodel_classes[field_type.__name__] = field_type
        else:
            field_type = new_basemodel_classes[field_type.__name__]
    # It's still possible there's a TypedDict in the args to be converted - e.g
    # Dict[str, SomeTypedDict]
    else:
        origin = get_origin(field_type)
        args = get_args(field_type)
        if origin and args:
            field_type = origin[
                tuple(
                    change_sub_typed_dicts_to_basemodels(arg, new_basemodel_classes)
                    for arg in args
                )
            ]
    return field_type


def field_parser(
    field_name: str, field_type: type, new_basemodel_classes: Dict[str, BaseModel]
):
    """Parses a field annotation and generates a pydantic friendly one by
    extracting relevant information.

    Parameters
    ----------------
    field_name : str
        Name of the field being parsed.
    field_type : type
        Annotation to be parsed.
    new_basemodel_classes : dict
        Dict where keys are names of _TypedDictMeta classes referenced inside
        the TypedDict being converted to a BaseModel, and values are BaseModel
        classes that have already been generated from those TypedDicts -
        pydantic will throw an error if we try to generate the same BaseModel
        twice so we just reference the one already created if multiple
        annotations have the same TypedDict.

    Returns
    -----------------
    field_info : type
        The field_type inputted, but parsed to be pydantic friendly
    """
    field_type, is_not_required = field_is_not_required(
        field_type, remove_origin_if_NotRequired=True
    )

    as_ref, field_info = get_annotation_contents(field_type)

    field_type = get_field_type(
        field_name,
        field_type,
        new_basemodel_classes,
    )

    if as_ref:
        field_type = parse_AsRef(field_type, field_info, as_ref, new_basemodel_classes)
        field_info.regex = None

    if is_not_required:
        field_type = Optional[field_type]

    return Annotated[field_type, field_info]


class Config(BaseConfig):
    """Config for generated BaseModel."""

    def alias_generator(string_to_be_aliased):
        """Alias in snake case"""
        return re.sub(r"(?<!^)(?=[A-Z])", "_", string_to_be_aliased).lower()


def strip_newline_literal(schema: dict):
    """
    Pydantic formats the docstring newlines by literally including `\n` in the
    description. To avoid this uglyness this function swaps `\n` for ` ` in
    every "description` key in the schema.
    """
    for key in schema:
        if isinstance(schema[key], dict):
            schema[key] = strip_newline_literal(schema[key])
        elif key == "description":
            schema[key] = schema[key].replace("\n", " ")
    return schema


def sort_alphabetically(schema: dict) -> dict:
    """Sorts the schema alphabetically by key name, exchanging the
    properties dicts for OrderedDicts"""
    schema = OrderedDict(sorted(list(schema.items()), key=lambda x: x[0]))

    return schema


def sort_schema(schema: dict) -> dict:
    if isinstance(schema, dict):
        schema = OrderedDict(
            sorted(
                list(schema.items()), key=lambda x: SortOrder.get(x[0], len(SortOrder))
            )
        )

        for key in schema:
            if key in ("definitions", "properties", "required"):
                if isinstance(schema[key], dict):
                    schema[key] = sort_alphabetically(schema[key])
                    for key2 in schema[key]:
                        if isinstance(schema[key][key2], dict):
                            schema[key][key2] = sort_schema(schema[key][key2])
                elif isinstance(schema[key], list):
                    schema[key].sort()

    return schema


SortOrder = {
    "title": 0,
    "description": 1,
    "type": 2,
    "definitions": 3,
    "properties": 4,
    "required": 5,
    "additionalProperties": 6,
    "patternProperties": 7,
}


# From https://github.com/pydantic/pydantic/issues/760#issuecomment-589708485
def parse_typeddict_to_schema(
    typed_dict: _TypedDictMeta,
    out_dir: Optional[Path] = None,
    return_basemodel: bool = False,
    new_basemodel_classes: Optional[Dict[str, BaseModel]] = None,
    sort: bool = True,
) -> Union[Type[BaseModel], Dict[str, type]]:
    """Takes a TypedDict and generates a jsonschema from it.

    Parameters
    ----------
    typed_dict : _TypedDictMeta
        The typeddict to be converted to a pydantic basemodel.
    out_dir: Optional[Path]
        Optionally provide a directory to store the generated json schema from
        the basemodel, if None then the dictionary schema won't be saved
        to disk.
    return_basemodel : bool
        Optionally return the basemodel as soon as it's generated, rather than
        converting it to a dictionary. Required for converting TypedDicts
        within documents.
    new_basemodel_classes : Dict[str, BaseModel]
        Optionally provide basemodel classes already generated during a
        conversion. Required for when the function is called recursively.
    sort : bool
        If true, sort the properties keys in the outputted schema.

    Returns
    -------
    Either the generated BaseModel or the schema dictionary generated from it,
    depending on if return_basemodel is True.
    """

    if new_basemodel_classes is None:
        new_basemodel_classes = {}
    annotations: Dict[str, type] = {}

    typed_dict = format_all_annotations(
        typed_dict, new_basemodel_classes=new_basemodel_classes
    )

    for name, field in typed_dict.__annotations__.items():
        default_value = getattr(typed_dict, name, ...)
        annotations[name] = (field, default_value)

    model = create_model(typed_dict.__name__, __config__=Config, **annotations)

    # Docstring is used as the description field.
    model.__doc__ = typed_dict.__doc__

    if return_basemodel:
        return model

    # title goes to snake_case
    model.__name__ = Config.alias_generator(typed_dict.__name__).lower()

    model_schema = model.schema(by_alias=True)
    model_schema = strip_newline_literal(model_schema)

    # Add the manually defined extra stuff
    if typed_dict in extra_schema:
        model_schema = merge_dicts(extra_schema[typed_dict], model_schema)

    if sort:
        model_schema = sort_schema(model_schema)

    if out_dir:
        with open(out_dir / f'{model_schema["title"]}.json', "w+") as f:
            json.dump(model_schema, f, indent=3)

    return model_schema


def generate_all_schema(schema_out_dir: Path = SCHEMA_OUT_DIR) -> None:
    """Generates all schema in the documents directory."""
    parse_typeddict_to_schema(DatumPage, out_dir=schema_out_dir)
    parse_typeddict_to_schema(Datum, out_dir=schema_out_dir)
    parse_typeddict_to_schema(EventDescriptor, out_dir=schema_out_dir)
    parse_typeddict_to_schema(EventPage, out_dir=schema_out_dir)
    parse_typeddict_to_schema(Event, out_dir=schema_out_dir)
    parse_typeddict_to_schema(Resource, out_dir=schema_out_dir)
    parse_typeddict_to_schema(RunStart, out_dir=schema_out_dir)
    parse_typeddict_to_schema(RunStop, out_dir=schema_out_dir)
    parse_typeddict_to_schema(StreamDatum, out_dir=schema_out_dir)
    parse_typeddict_to_schema(StreamResource, out_dir=schema_out_dir)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--schema_out_directory", default=SCHEMA_OUT_DIR, nargs="?")
    args = parser.parse_args()
    generate_all_schema(schema_out_dir=args.schema_out_directory)



from typing import Any, Dict, List

from typing_extensions import Annotated, TypedDict

from .generate.type_wrapper import AsRef, Field, add_extra_schema

DATUM_PAGE_EXTRA_SCHEMA = {"additionalProperties": False}


@add_extra_schema(DATUM_PAGE_EXTRA_SCHEMA)
class DatumPage(TypedDict):
    """Page of documents to reference a quanta of externally-stored data"""

    datum_id: Annotated[
        List[str],
        AsRef("Dataframe"),
        Field(
            description="Array unique identifiers for each Datum (akin to 'uid' for "
            "other Document types), typically formatted as '<resource>/<integer>'"
        ),
    ]
    datum_kwargs: Annotated[
        Dict[str, List[Any]],
        Field(
            description="Array of arguments to pass to the Handler to "
            "retrieve one quanta of data"
        ),
    ]
    resource: Annotated[
        str,
        Field(
            description="The UID of the Resource to which all Datums in the page belong"
        ),
    ]
